/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.common;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import kieker.analysis.AnalysisController;
import kieker.analysis.AnalysisController.STATE;
import kieker.analysis.display.AbstractDisplay;
import kieker.analysis.exception.AnalysisConfigurationException;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.webgui.common.exception.AnalysisNotInstantiatedException;
import kieker.webgui.common.exception.AnalysisNotRunningException;
import kieker.webgui.common.exception.ProjectAlreadyStartedException;
import kieker.webgui.common.exception.ProjectStillRunningException;

/**
 * This manager is responsible for the currently used and running instances of {@code AnalysisController}. It supplies methods to check the states of analysis,
 * instantiate, start and to stop them. This is also a singleton instance.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
public final class ACManager {
	/**
	 * This is the maximal time the application will wait for an analysis thread to be terminated.
	 */
	private static final long MAX_THREAD_WAIT_TIME_MS = 1000;
	/**
	 * This is the singleton instance of this class.
	 */
	private static final ACManager INSTANCE = new ACManager();
	/**
	 * This list contains the current analysis controllers and their corresponding threads. Not every project does have a controller, but every project can have
	 * maximal one.
	 */
	private final ConcurrentHashMap<String, Pair<AnalysisController, Thread>> analysisController = new ConcurrentHashMap<String, Pair<AnalysisController, Thread>>();

	/**
	 * Creates a new instance of this class.
	 */
	private ACManager() {
		// No code necessary.
	}

	/**
	 * Delivers the singleton instance of this class.
	 * 
	 * @return The only instance of {@link FSManager}.
	 */
	public static ACManager getInstance() {
		return ACManager.INSTANCE;
	}

	/**
	 * This method tries to stop a running analysis.
	 * 
	 * @param project
	 *            The project to be stopped.
	 * @throws AnalysisNotRunningException
	 *             If the analysis is not running or has not yet been initialized.
	 * @throws InterruptedException
	 *             If the current thread has somehow been interrupted while waiting for the existing analysis thread.
	 */
	public void stopAnalysisController(final String project) throws AnalysisNotRunningException, InterruptedException {
		final Pair<AnalysisController, Thread> currController = this.analysisController.get(project);
		synchronized (currController) {
			// Is there a thread available?
			if ((currController.getFst() == null) || (currController.getSnd() == null)) {
				throw new AnalysisNotRunningException();
			}

			// Try to stop the analysis
			currController.getFst().terminate();
			currController.getSnd().join(ACManager.MAX_THREAD_WAIT_TIME_MS);
		}
	}

	/**
	 * This method tries to start the analysis for the given project. It has already to be initialized.
	 * 
	 * @param project
	 *            The project to be started.
	 * @throws ProjectAlreadyStartedException
	 *             If the project is already running.
	 * @throws AnalysisNotInstantiatedException
	 *             If the analysis has not yet been initialized.
	 */
	public void startAnalysisController(final String project) throws ProjectAlreadyStartedException, AnalysisNotInstantiatedException {
		final Pair<AnalysisController, Thread> currController = this.analysisController.get(project);
		synchronized (currController) {
			// Is there an analysis after all?
			if (currController.getFst() == null) {
				throw new AnalysisNotInstantiatedException();
			}
			// Check whether this analysis has already been started.
			if (currController.getSnd() != null) {
				throw new ProjectAlreadyStartedException();
			} else {
				// Otherwise start a thread and run the analysis in it.
				currController.setSnd(new Thread() {
					@Override
					public void run() {
						try {
							currController.getFst().run();
						} catch (final IllegalStateException ex) {
							ex.printStackTrace();
						} catch (final AnalysisConfigurationException ex) {
							ex.printStackTrace();
						}
					}
				});

				currController.getSnd().start();
			}
		}
	}

	/**
	 * This method tries to instantiate the analysis controller of the given project using the data available on the file system.
	 * 
	 * @param project
	 *            The project whose analysis controller should be instantiated.
	 * @throws NullPointerException
	 *             If an error occurred while loading the project.
	 * @throws AnalysisConfigurationException
	 *             If the given configuration is somehow invalid.
	 * @throws IOException
	 *             If an error occurred while loading the project.
	 * @throws ProjectStillRunningException
	 *             If there exists already an analysis controller and is still running.
	 */
	public void instantiateAnalysisController(final String project) throws NullPointerException, AnalysisConfigurationException, IOException,
			ProjectStillRunningException {
		// Add the potential new project atomically
		final MIProject modelProject = FSManager.getInstance().openProject(project);
		final ClassLoader classLoader = FSManager.getInstance().getClassLoader(project);
		final Pair<AnalysisController, Thread> newController = new Pair<AnalysisController, Thread>(new AnalysisController(modelProject, classLoader), null);
		final Pair<AnalysisController, Thread> currController = this.analysisController.putIfAbsent(project, newController);

		if (currController != null) {
			// There is currently an analysis controller available. Decide what to do.
			synchronized (currController) {
				switch (currController.getFst().getState()) {
				case FAILED:
				case TERMINATED:
				case READY:
					// No problem in instantiating a new controller
					this.analysisController.put(project, newController);
					break;
				case RUNNING:
					throw new ProjectStillRunningException("The project with the name '" + project + "' is still running.");
				default:
					// No code necessary
					break;
				}
			}
		}
	}

	/**
	 * This method can be used to deliver the state of the analysis controller of the given project as a human readable string.
	 * 
	 * @param project
	 *            The project whose state should be delivered.
	 * @return The current state of the corresponding {@link AnalysisController} if it exists, '' otherwise.
	 */
	public String getAnalysisControllerStateString(final String project) {
		final STATE controllerState = this.getAnalysisControllerState(project);
		final String controllerStateString;

		if (controllerState == null) {
			controllerStateString = "";
		} else {
			controllerStateString = controllerState.toString();
		}

		return controllerStateString;
	}

	/**
	 * This method can be used to deliver the state of the analysis controller of the given project as a human readable string.
	 * 
	 * @param project
	 *            The project whose state should be delivered.
	 * @return The current state of the corresponding {@link AnalysisController} if it exists, null otherwise.
	 */
	public STATE getAnalysisControllerState(final String project) {
		// If the given project is null, we return a null-state.
		if (project == null) {
			return null;
		}
		final Pair<AnalysisController, Thread> controller = this.analysisController.get(project);

		final STATE controllerState;
		if (controller == null) {
			controllerState = null;
		} else {
			synchronized (controller) {
				if (controller.getFst() == null) {
					controllerState = null;
				} else {
					controllerState = controller.getFst().getState();
				}
			}
		}

		return controllerState;
	}

	/**
	 * This method delivers the current displays of the given project. They can be used to show the stored information but should <b>not</b> be modified.
	 * 
	 * @param projectName
	 *            The name of the project whose displays should be delivered.
	 * @return A list of pairs containing the displays and the name of them.
	 */
	public List<Pair<String, AbstractDisplay>> getCurrentDisplays(final String projectName) {
		final List<Pair<String, AbstractDisplay>> result = new ArrayList<Pair<String, AbstractDisplay>>();

		return result;
	}

	/**
	 * This helper thread is used to update the available displays of the ACManager at regular intervals. <b>Important:</b> If the interrupt-method of the thread is
	 * being called, it will be terminated.
	 * 
	 * @author Nils Christian Ehmke
	 * @version 1.0
	 */
	@SuppressWarnings("unused")
	private static class DisplayUpdateThread extends Thread {

		/**
		 * This is the time the thread waits between the updates.
		 */
		private static final long SLEEP_TIME_MS = 2 * 1000;

		/**
		 * Default constructor.
		 */
		public DisplayUpdateThread() {
			// No code necessary
		}

		@Override
		public void run() {
			// Run until we have been interrupted
			while (!Thread.interrupted()) {
				// TODO Implement

				// Wait a little bit.
				try {
					Thread.sleep(DisplayUpdateThread.SLEEP_TIME_MS);
				} catch (final InterruptedException ex) {
					// We have been interrupted. Exit the thread
					return;
				}
			}
		}
	}
}
